/* Sysmon - dockapp system monitor kernel module support code
 * Copyright (C) 2000 timecop@japan.co.jp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/miscdevice.h>
#include <linux/kernel_stat.h>
#include <linux/mm.h>
/* #include <linux/pagemap.h> */
/* #include <linux/swap.h> */
/* #include <asm/io.h> */
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <linux/types.h>
/* #include <linux/fs.h> */
#include <linux/version.h>

#define SYSMON_MINOR	253
#define SMIOC_GETKSTAT	100
#define SMIOC_GETSI	101

MODULE_DESCRIPTION("Module to access kstat structure from userspace");
MODULE_AUTHOR("timecop [timecop@japan.co.jp]");

static int sysmon_open(struct inode *inode, struct file *file)
{
    MOD_INC_USE_COUNT;
    return 0;
}

static int sysmon_release(struct inode *inode, struct file *file)
{
    MOD_DEC_USE_COUNT;
    return 0;
}

static int sysmon_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
    switch (cmd) {
    case SMIOC_GETKSTAT:
	/* we don't need to waste time doing verify_area
	 * this only has one use, and kstat is already allocated */
	return copy_to_user((struct kernel_stat *) arg, &kstat, sizeof(kstat));
    case SMIOC_GETSI:
	return -ENOIOCTLCMD;
	break;
    default:
	return -ENOIOCTLCMD;
    }
}

struct file_operations sysmon_fops = {
    open:sysmon_open,
    release:sysmon_release,
    ioctl:sysmon_ioctl
};

static struct miscdevice sysmon_dev = {
    SYSMON_MINOR,
    "sysmon",
    &sysmon_fops
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
int init_module(void)
#else
static int __init sysmon_init(void)
#endif
{
    misc_register(&sysmon_dev);
    printk(KERN_INFO "sysmon module loaded\n");
    return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
void cleanup_module(void)
#else
static void __exit sysmon_cleanup(void)
#endif
{
    misc_deregister(&sysmon_dev);
    printk(KERN_INFO "sysmon module unloaded\n");
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
module_init(sysmon_init);
module_exit(sysmon_cleanup);
#endif
