/* Sysmon - dockapp system monitor
 * Copyright (C) 2000 timecop@japan.co.jp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#if defined(EQ) && defined(MODE5)
#error Cannot define EQ and MODE5 at the same time!
#endif
#if defined(EQ) && defined(HIGHIRQ)
#error Cannot define EQ and HIGHIRQ at the same time!
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/utsname.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <getopt.h>
#include <syscall.h>
#include <errno.h>
#include <linux/kernel_stat.h>
#include <linux/kernel.h>
#include <linux/version.h>

/* x11 includes */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include <X11/xpm.h>

/* pixmaps */
#include "master.xpm"

/* UI defines for offsets inside pixmap and things like that. */
#ifdef HIGHIRQ
#define BAR_FULL	5
#define BAR_HALF	3
#else
#define BAR_FULL	6
#define BAR_HALF	3
#endif

/* lights and bars */
#define MODE_RED	0
#define MODE_GREEN	4
#define MODE_OFF	8

/* locations of memory, rio, wio bars. prepare for 24 IRQ setups */
#ifdef HIGHIRQ
#define BAR_MEM_PFY	3, 4
#define BAR_MEM_SWP	3, 13
#define BAR_IO_READ	3, 22
#define BAR_IO_WRIT	3, 25
#else
#define BAR_MEM_PFY	3, 4
#define BAR_MEM_SWP	3, 15
#define BAR_IO_READ	3, 26
#define BAR_IO_WRIT	3, 29
#endif
/* locations of swap and page lights */
#define LGT_PGINX	51
#define LGT_PGINY	51
#define LGT_PGOUTX	56
#define LGT_PGOUTY	51
#define LGT_SWPINX	51
#define LGT_SWPINY	56
#define LGT_SWPOUTX	56
#define LGT_SWPOUTY	56

/* locations of uptime line */
#ifdef HIGHIRQ
#define UPTIME_1	33
#else
#define UPTIME_1	38
#endif

/* sysmon ioctl */
#define SMIOC_GETKSTAT	100
#define SMIOC_GETSI	101	/* unused atm */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
#error if (KERNEL_VERSION == 2.0) user = moron;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
#define DK_MAX_DISK DK_NDRIVE
#endif

typedef void (*function) (void);

typedef struct {
    Display *d;
    int screen;
    Window root;
    Window win;
#ifndef BLACKBOX
    Window iconwin;
#endif
    Pixmap pixmap;
    Pixmap mask;
    GC gc;
    int width;
    int height;
#ifdef TWM
    int x;
    int y;
#endif
    time_t uptime;
    time_t time;
} Dockapp;

/* X11 support functions */
void flush_expose(void);
void new_window(char *, int, int);
#ifndef BLACKBOX
#define copy_xpm_area(x, y, w, h, dx, dy)				\
    XCopyArea(dockapp.d, dockapp.pixmap, dockapp.pixmap, dockapp.gc,	\
	      x, y, w, h, dx, dy);					\
    XCopyArea(dockapp.d, dockapp.pixmap, dockapp.iconwin, dockapp.gc,	\
	      x, y, w, h, dx, dy);
#else
#define copy_xpm_area(x, y, w, h, dx, dy)				\
    XCopyArea(dockapp.d, dockapp.pixmap, dockapp.pixmap, dockapp.gc,	\
	      x, y, w, h, dx, dy);					\
    XCopyArea(dockapp.d, dockapp.pixmap, dockapp.win, dockapp.gc,	\
	      x, y, w, h, dx, dy);
#endif

/* draw support functions */
#ifndef CONFIG_SMP
    /* standard one-cpu copy */
#define irq_prep()							\
    register int j;							\
    for (j = 0; j < smp_num_cpus; j++) {				\
	my_memcpy(irq, kstat.irqs[cpu_logical_map(j)], sizeof(irq));	\
    }
#else
    /* SMP use.  total all the IRQs */
#define irq_prep()							\
    register int j,j1;							\
    memset(irq, 0, sizeof(irq));					\
    for (j = 0; j < smp_num_cpus; j++)					\
	for (j1 = 0; j1 < IRQN; j1++)					\
	    irq[j1] = irq[j1] + kstat.irqs[cpu_logical_map(j)][j1];
#endif

static void draw_bar(int x, int y, int width, int mode);
static void do_irq_bar();
static void do_irq_led();
/* draw functions */
static void draw_thread(function irq_fn, int mem);
static void uptime_thread(void);
/* support */
#if defined(MMX) && defined(__i386__)
/* directly from Linux kernel. In my program we copy either 64 or 128 bytes
 * depending on HIGHIRQ set or not.  Is this really faster?  Dunno.  It's
 * here, and if it works for you, that's good.  Also, since we are always
 * aligned on 64, got rid of the normal memcpy near the end. Hell, this can
 * potentially kill your system, delete all your files, and cause penis
 * shrinkage.  But I take NO RESPONSIBILITY for whatever happens to you if
 * this code is executed. */
void *my_memcpy(void *to, const void *from, size_t len)
{
/* *INDENT-OFF* */ 
    void *p = to;
    int i = len >> 6;
    __asm__ __volatile__ (
	"1: prefetch (%0)\n"
	"	prefetch 64(%0)\n"
	"	prefetch 128(%0)\n"
	"	prefetch 192(%0)\n"
	"	prefetch 256(%0)\n"
	"	2:  \n"
	".section .fixup, \"ax\"\n"
	"3: movw $0x1AEB, 1b\n"
	"	jmp 2b\n"
	".previous\n"
	".section __ex_table,\"a\"\n"
	"	.align 4\n"
	"	.long 1b, 3b\n"
	".previous"
	: : "r" (from)
    );
		
    for(; i > 0; i--) {
	__asm__ __volatile__ (
	    "1:  prefetch 320(%0)\n"
	    "	2:  movq (%0), %%mm0\n"
	    "	movq 8(%0), %%mm1\n"
	    "	movq 16(%0), %%mm2\n"
	    "	movq 24(%0), %%mm3\n"
	    "	movq %%mm0, (%1)\n"
	    "	movq %%mm1, 8(%1)\n"
	    "	movq %%mm2, 16(%1)\n"
	    "	movq %%mm3, 24(%1)\n"
	    "	movq 32(%0), %%mm0\n"
	    "	movq 40(%0), %%mm1\n"
	    "	movq 48(%0), %%mm2\n"
	    "	movq 56(%0), %%mm3\n"
	    "	movq %%mm0, 32(%1)\n"
	    "	movq %%mm1, 40(%1)\n"
	    "	movq %%mm2, 48(%1)\n"
	    "	movq %%mm3, 56(%1)\n"
	    ".section .fixup, \"ax\"\n"
	    "	3: movw $0x05EB, 1b\n"
	    "	jmp 2b\n"
	    ".previous\n"
	    ".section __ex_table,\"a\"\n"
	    "	.align 4\n"
	    "	.long 1b, 3b\n"
	    ".previous"
	    : : "r" (from), "r" (to) : "memory"
	);
	from+=64;
	to+=64;
    }
    return p;
    /* because I aligned the arrays so that this would always get called with
     * size_t size of 0, instead of wasting time calling it, let's not. */
    /* memcpy(to, from, len&63); */
/* *INDENT-ON* */
}

#else
#define my_memcpy memcpy
#endif
static void catch_alarm(int sig);
void kernel_check(void);
void sysmon_help(void);

/* globals */
static Dockapp dockapp;

#ifdef EQ
int pmode[10] = { 92, 90, 88, 86, 84, 82, 80, 78, 76, 74 };
#endif
#if defined(MODE5)
int pmode[5] = { 28, 24, 20, 16, 12 };
#endif
#if !defined(EQ) && !defined(MODE5)
int pmode[4] = { 24, 20, 16, 12 };
#endif

#if defined(EQ)
#define IRQN	16
int xpos[16] =
    { 4, 6, 9, 11, 14, 16, 19, 21, 24, 26, 29, 31, 34, 36, 39, 41 };
int ypos[16] =
    { 51, 51, 51, 51, 51, 51, 51, 51, 51, 51, 51, 51, 51, 51, 51, 51 };
#endif

#if defined(HIGHIRQ)
#define IRQN	32
int xpos[32] =
    { 4, 9, 14, 19, 24, 29, 34, 39, 4, 9, 14, 19, 24, 29, 34, 39, 4, 9, 14,
    19, 24, 29, 34, 39, 0, 0, 0, 0, 0, 0, 0, 0
};
int ypos[32] =
    { 46, 46, 46, 46, 46, 46, 46, 46, 51, 51, 51, 51, 51, 51, 51, 51, 56,
    56, 56, 56, 56, 56, 56, 56, 0, 0, 0, 0, 0, 0, 0, 0
};
#endif

#if !defined(EQ) && !defined(HIGHIRQ)
#define IRQN	16
int xpos[16] =
    { 4, 9, 14, 19, 24, 29, 34, 39, 4, 9, 14, 19, 24, 29, 34, 39 };
int ypos[16] =
    { 51, 51, 51, 51, 51, 51, 51, 51, 56, 56, 56, 56, 56, 56, 56, 56 };
#endif

struct kernel_stat kstat;	/* where all the magic happens */
struct sysinfo si;		/* ditto */
unsigned long int irq[IRQN];	/* system IRQ table. */
unsigned int irqpeak[IRQN];	/* this tries to keep insane peaks from updating */
int fd;				/* sysmon fd for ioctl calls */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
int shift;			/* amount to bitshift to calculate some memory stuff */
#endif
int refresh_rate = 100000;	/* refresh rate. modified by the -r param */

/* signal catcher for SIGALRM. Calling this will reset the alarm and update
 * uptime display.  So, uptime display can be up to 59 seconds off, depending
 * when you started tcsysmon.  I don't think that's such a big deal though. */
static void catch_alarm(int sig)
{
    signal(sig, catch_alarm);
    uptime_thread();
}

/* UI */
void flush_expose(void)
{
    XEvent dummy;
#ifndef BLACKBOX
    while (XCheckTypedWindowEvent
	   (dockapp.d, dockapp.iconwin, Expose, &dummy));
#else
    while (XCheckTypedWindowEvent(dockapp.d, dockapp.win, Expose, &dummy));
#endif
}

void new_window(char *name, int width, int height)
{
    XpmAttributes attr;
    Pixel fg, bg;
    XGCValues gcval;
#ifdef HIGHIRQ
    GC temp;
#endif
    XSizeHints sizehints;
    XClassHint classhint;
    XWMHints wmhints;
#if defined(TWM) && defined(BLACKBOX)
    XSetWindowAttributes xattr;
#endif

    dockapp.width = width;
    dockapp.height = height;
    dockapp.screen = DefaultScreen(dockapp.d);
    dockapp.root = DefaultRootWindow(dockapp.d);

    sizehints.flags = USSize | USPosition;
#ifdef TWM
    sizehints.x = dockapp.x ? dockapp.x : 0;
    sizehints.y = dockapp.y ? dockapp.y : 0;
#else
    sizehints.x = 0;
    sizehints.y = 0;
#endif
    sizehints.width = width;
    sizehints.height = height;

    fg = BlackPixel(dockapp.d, dockapp.screen);
    bg = WhitePixel(dockapp.d, dockapp.screen);

    dockapp.win = XCreateSimpleWindow(dockapp.d, dockapp.root,
				      sizehints.x, sizehints.y,
				      sizehints.width, sizehints.height, 1,
				      fg, bg);
#ifndef BLACKBOX
    dockapp.iconwin = XCreateSimpleWindow(dockapp.d, dockapp.win,
					  sizehints.x, sizehints.y,
					  sizehints.width,
					  sizehints.height, 1, fg, bg);
#endif

    XSetWMNormalHints(dockapp.d, dockapp.win, &sizehints);
    classhint.res_name = name;
    classhint.res_class = name;
    XSetClassHint(dockapp.d, dockapp.win, &classhint);

    XSelectInput(dockapp.d, dockapp.win,
		 ExposureMask | StructureNotifyMask);
#ifndef BLACKBOX
    XSelectInput(dockapp.d, dockapp.iconwin,
		 ExposureMask | StructureNotifyMask);
#endif

    XStoreName(dockapp.d, dockapp.win, name);
    XSetIconName(dockapp.d, dockapp.win, name);

    gcval.foreground = fg;
    gcval.background = bg;
    gcval.graphics_exposures = False;

    dockapp.gc =
	XCreateGC(dockapp.d, dockapp.win,
		  GCForeground | GCBackground | GCGraphicsExposures,
		  &gcval);

    attr.exactColors = 0;
    attr.alloc_close_colors = 1;
    attr.closeness = 1L << 15;
    attr.valuemask = (XpmExactColors | XpmAllocCloseColors | XpmCloseness);
    if (XpmCreatePixmapFromData(dockapp.d, dockapp.win,
				master_xpm, &dockapp.pixmap, &dockapp.mask,
				&attr) != XpmSuccess) {
	fprintf(stderr, "Not enough colors!\n");
	exit(-1);
    }
#ifdef HIGHIRQ
    XCopyArea(dockapp.d, dockapp.pixmap, dockapp.pixmap, dockapp.gc, 64, 0,
	      dockapp.width, dockapp.height, 0, 0);
    temp =
	XCreateGC(dockapp.d, dockapp.mask,
		  GCForeground | GCBackground | GCGraphicsExposures,
		  &gcval);
    XCopyArea(dockapp.d, dockapp.mask, dockapp.mask, temp, 64, 0,
	      dockapp.width, dockapp.height, 0, 0);
#endif

    XShapeCombineMask(dockapp.d, dockapp.win, ShapeBounding, 0, 0,
		      dockapp.mask, ShapeSet);
#ifndef BLACKBOX
    XShapeCombineMask(dockapp.d, dockapp.iconwin, ShapeBounding, 0, 0,
		      dockapp.mask, ShapeSet);
#endif

    wmhints.initial_state = WithdrawnState;
    wmhints.flags = StateHint;
#ifndef BLACKBOX
    wmhints.icon_window = dockapp.iconwin;
    wmhints.icon_x = sizehints.x;
    wmhints.icon_y = sizehints.y;
    wmhints.window_group = dockapp.win;
    wmhints.flags =
	StateHint | IconWindowHint | IconPositionHint | WindowGroupHint;
#endif
    XSetWMHints(dockapp.d, dockapp.win, &wmhints);
#if defined(TWM) && defined(BLACKBOX)
    xattr.override_redirect = True;
    XChangeWindowAttributes(dockapp.d, dockapp.win, CWOverrideRedirect,
			    &xattr);
#endif
    XMapWindow(dockapp.d, dockapp.win);
}

static void draw_bar(int x, int y, int width, int mode)
{
    copy_xpm_area(0, 74, width, mode, x, y);
    copy_xpm_area(0, 81, 58 - width, mode, x + width, y);
}

static void do_irq_bar(void)
{
    static unsigned long int lints[IRQN];
    static int peak[IRQN];
    register int intdiff;
    register int i;
    register int pos;
    irq_prep();
    /* no more irqstat ioctl, this has been provided in kstat all along */

    /* draw interrupt bars.  peak calculation code was taken
     * from wmsysmon 0.6.0. see readme for copyright info */
    for (i = 0; i < IRQN; i++) {
	pos = 0;
	intdiff = irq[i] - lints[i];
	/* did you know the next line divides by 2? */
	peak[i] = (intdiff + peak[i]) >> 1;
	if (peak[i]) {
	    if (intdiff) {
#ifdef EQ
		pos = intdiff * 10 / peak[i];
#endif
#ifdef MODE5
		pos = (intdiff << 2) / peak[i];
#endif
#if !defined(MODE5) && !defined(EQ)
		pos = intdiff * 3 / peak[i];
#endif
	    } else {
		pos = peak[i];
	    }
	}
#ifdef EQ
	if (pos > 9)
	    pos = 9;
#endif
#ifdef MODE5
	if (pos > 4)
	    pos = 4;
#endif
#if !defined(EQ) && !defined(MODE5)
	if (pos > 3)
	    pos = 3;
#endif
	if (irqpeak[i] != pos) {
#ifdef EQ
	    copy_xpm_area(pmode[pos], 74, 2, 9, xpos[i], ypos[i]);
#else
	    copy_xpm_area(pmode[pos], 88, 4, 4, xpos[i], ypos[i]);
#endif
	    irqpeak[i] = pos;
	}
    }
    my_memcpy(lints, irq, sizeof(irq));
}

static void do_irq_led(void)
{
    static unsigned long int lints[IRQN];
    static unsigned long int lmask;
    register int i;
    register int tmp;
    register int tmy;
    irq_prep();
    /* no more irqstat ioctl, this has been provided in kstat all along */

    /* draw interrupt lights */
    for (i = 0; i < IRQN; i++) {
	tmp = irq[i] != lints[i];
	tmy = lmask & (1 << i);
	if (tmp && !tmy) {
	    copy_xpm_area(MODE_GREEN, 88, 4, 4, xpos[i], ypos[i]);
	    lmask |= (1 << i);
	} else if (!tmp && tmy) {
	    copy_xpm_area(MODE_OFF, 88, 4, 4, xpos[i], ypos[i]);
	    lmask ^= (1 << i);
	}
    }
    my_memcpy(lints, irq, sizeof(irq));
}

/* after some thought, moving all the updates into one big function wasn't
 * such a bad idea.  Before the shit was separate because it read from 3
 * different fd's, but now it just does 2 ioctl's so, might as well keep
 * them all here, especially since some called the ioctl more than once. */
static void draw_thread(function irq_fn, int mem)
{
    /* general */
    static int count;

    /* swap, io */
    static unsigned long int oldpin, oldpout, oldsin, oldsout;
    static unsigned short int p[4];
    static unsigned short int r[2];
#ifndef IO_MAJOR
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
    register int major;
#endif
#endif
    register int disk;

    /* interrupts + io */
    long int read_io = 0, write_io = 0;
    static long int old_read_io, old_write_io;

    /* memory */
    char *ptr;
    unsigned long cached;
    unsigned int mbar, sbar;
    unsigned long int diff;
    static unsigned int oldmem, oldswap;

    /* get system info */
    ioctl(fd, SMIOC_GETKSTAT, &kstat);

    /* draw IRQ lights or bars */
#ifdef EQ
    /* bars only */
    do_irq_bar();
#else
    /* choose */
    (*irq_fn) ();
#endif

    /* draw IO and swap */
    /* kernel 2.2 doesnt have anything fancy.  just total'em up. */
#ifndef IO_MAJOR
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
    for (disk = 0; disk < DK_MAX_DISK; disk++) {
	read_io += kstat.dk_drive_rio[disk];
	write_io += kstat.dk_drive_wio[disk];
    }
#else
    for (major = 0; major < DK_MAX_MAJOR; major++) {
	for (disk = 0; disk < DK_MAX_DISK; disk++) {
	    int active = kstat.dk_drive_rio[major][disk] +
		kstat.dk_drive_wio[major][disk];
	    if (active) {
		read_io += kstat.dk_drive_rio[major][disk];
		write_io += kstat.dk_drive_wio[major][disk];
	    }
	}
    }
#endif
#else
    for (disk = 0; disk < DK_MAX_DISK; disk++) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	read_io += kstat.dk_drive_rio[disk];
	write_io += kstat.dk_drive_wio[disk];
#else
	read_io += kstat.dk_drive_rio[IO_MAJOR][disk];
	write_io += kstat.dk_drive_wio[IO_MAJOR][disk];
#endif
    }
#endif

    if (read_io > old_read_io) {
	diff = read_io - old_read_io;
	if (diff > 58)
	    diff = 58;
	draw_bar(BAR_IO_READ, diff, BAR_HALF);
	r[0] = 0;
	old_read_io = read_io ? read_io : old_read_io;
    } else if (!r[0]) {
	draw_bar(BAR_IO_READ, 0, BAR_HALF);
	r[0] = 1;
    }

    if (write_io > old_write_io) {
	diff = write_io - old_write_io;
	if (diff > 58)
	    diff = 58;
	draw_bar(BAR_IO_WRIT, diff, BAR_HALF);
	r[1] = 0;
	old_write_io = write_io ? write_io : old_write_io;
    } else if (!r[1]) {
	draw_bar(BAR_IO_WRIT, 0, BAR_HALF);
	r[1] = 1;
    }

    /* swap, page in/out */
    if (kstat.pgpgin > oldpin) {
	copy_xpm_area(MODE_RED, 88, 4, 4, LGT_PGINX, LGT_PGINY);
	p[0] = 0;
	oldpin = kstat.pgpgin;
    } else if (!p[0]) {
	copy_xpm_area(MODE_OFF, 88, 4, 4, LGT_PGINX, LGT_PGINY);
	p[0] = 1;
    }

    if (kstat.pgpgout > oldpout) {
	copy_xpm_area(MODE_RED, 88, 4, 4, LGT_PGOUTX, LGT_PGOUTY);
	p[1] = 0;
	oldpout = kstat.pgpgout;
    } else if (!p[1]) {
	copy_xpm_area(MODE_OFF, 88, 4, 4, LGT_PGOUTX, LGT_PGOUTY);
	p[1] = 1;
    }

    if (kstat.pswpin != oldsin) {
	copy_xpm_area(MODE_RED, 88, 4, 4, LGT_SWPINX, LGT_SWPINY);
	p[2] = 0;
	oldsin = kstat.pswpin;
    } else if (!p[2]) {
	copy_xpm_area(MODE_OFF, 88, 4, 4, LGT_SWPINX, LGT_SWPINY);
	p[2] = 1;
    }

    if (kstat.pswpout != oldsout) {
	copy_xpm_area(MODE_RED, 88, 4, 4, LGT_SWPOUTX, LGT_SWPOUTY);
	p[3] = 0;
	oldsout = kstat.pswpout;
    } else if (!p[3]) {
	copy_xpm_area(MODE_OFF, 88, 4, 4, LGT_SWPOUTX, LGT_SWPOUTY);
	p[3] = 1;
    }
    /* holy shit.  I ran this code on 2.2.10 and it was using fucking 8% CPU.
     * What actually happens, is under 2.2, si_meminfo walks through large
     * amounts of memory pages to see if they are reserved or not.  Since
     * proc driver calls si_meminfo to get memory information, it uses up
     * quite a bit of CPU time.  So, we will memory info every so often,
     * instead of every update_rate. Swapinfo checking is identical between
     * 2.2 and 2.4 so we don't care about that. Also the sysinfo syscall
     * does a cli() between uptime, nr_procs and load calculations. */
    if (--count < 0) {
	char buf[512];
	count = 30;
	syscall(__NR_sysinfo, &si);

	/* memory.  fuck, i wish kernel exported page_cache_size. I wouldn't have
	 * to do this retarded file IO in proc. by the way the stuff in bytes in
	 * proc/meminfo top part is supposed to go away as soon as "all the
	 * tools are updated". so this kB value is all we are going to get
	 * in the future. lets be prepared. I swear to god I will delete
	 * this so fast after finding out how to get memory info from the
	 * kernel */
	lseek(mem, 0L, SEEK_SET);
	read(mem, &buf, 512);
	ptr = strstr(buf, "Cache");
	/* ptr = strchr(buf, 'C'); either one works. Which is faster? */
	cached = atoi(ptr + 8);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	/* this crap overflows normal integer operations, so we bit
	 * shift it for fun. */
	mbar =
	    ((((si.totalram >> 10) - (si.freeram >> 10)) -
	      (si.bufferram >> 10) - cached) * 58) / (si.totalram >> 10);
#else
	mbar =
	    (((si.totalram - si.freeram) - si.bufferram -
	      ((cached << 10) >> shift)) * 58) / si.totalram;
#endif
	if (mbar != oldmem) {
	    draw_bar(BAR_MEM_PFY, mbar, BAR_FULL);
	    oldmem = mbar;
	}
	/* totalswap is incremented to avoid divide by 0 error. We don't want to
	 * do our math in floats though, so this is what we have to do. error
	 * introduced is within si.mem_unit, on i80x86, 4k. no problem.
	 * if there is any error introduced by this you won't notice until you swap
	 * is at 100% but by that time you got different kind of problems :)
	 * heheheheheheheeh */
	sbar = ((si.totalswap - si.freeswap) * 58) / ++si.totalswap;
	if (sbar != oldswap) {
	    draw_bar(BAR_MEM_SWP, sbar, BAR_FULL);
	    oldswap = sbar;
	}
    }
}

static void uptime_thread(void)
{
    time_t uptime;
    struct tm *upt;
    register char c;
    register char k = 5;
    char buf[10];
    char *ptr;

    alarm(60);
    time(&uptime);
    uptime = uptime - dockapp.time + dockapp.uptime;
    upt = gmtime(&uptime);
#ifdef _GNU_SOURCE
    snprintf(buf, 10, "%03d:%02d:%02d", upt->tm_yday, upt->tm_hour,
	     upt->tm_min);
#else
    sprintf(buf, "%03d:%02d:%02d", upt->tm_yday, upt->tm_hour,
	    upt->tm_min);
#endif
    ptr = buf;
    while (*ptr) {
	c = *ptr++ ^ 48;
	copy_xpm_area(c << 3, 65, 5, 7, k, UPTIME_1);
	k += 6;
    }
}

void sysmon_help(void)
{
    printf
	("tcsysmon 1.4 - by timecop@japan.co.jp [www.ne.jp/asahi/linux/timecop]\n"
	 "               original concept by dave clark [clarkd@skynet.ca]\n"
	 "               LED bar mode concept by swivel@gnugeneration.com\n"
	 "\n" "command line switches:\n" "\n"
	 "   -d display_name		connect to display_name\n"
#ifndef EQ
	 "   -l				Swivel mode (default off)\n"
#endif
	 "   -r rate			update rate (default 100)\n"
	 "   -h				show this help again\n"
	 "   -s	device			system monitor device file\n"
	 "				( usually /dev/sysmon (see README))\n"
#ifdef TWM
	 "   -x xpos			X-coordinate of the dockapp window\n"
	 "   -y ypos			Y-coordinate of the dockapp window\n"
#endif
	 "   -v				print version\n");
}

void kernel_check(void)
{

    int r, v, s;
    int r1, v1, s1;
    struct utsname un;

    uname(&un);
    sscanf(un.release, "%d.%d.%d", &r, &v, &s);
    sscanf(UTS_RELEASE, "%d.%d.%d", &r1, &v1, &s1);
    if ((r != r1) || (v != v1)) {
	printf("tcsysmon was compiled for %d.%d kernel.\n"
	       "Recompile is necessary to run it under %d.%d kernel.\n",
	       r1, v1, r, v);
	exit(-1);
    }

}

int main(int argc, char **argv)
{
    char *dispname = NULL;
    char *sysmon = NULL;
    int memory_fd;
#ifdef PRO
    int cnt = 0;
#endif
    int ch;
    XEvent event;
    function irq_fn = do_irq_led;

    while ((ch = getopt(argc, argv, "d:hlr:s:vx:y:")) != EOF) {
	switch (ch) {
	case 'd':
	    dispname = malloc(strlen(optarg) * sizeof(char));
	    strncpy(dispname, optarg, strlen(optarg));
	    break;
	case 'h':
	    sysmon_help();
	    exit(0);
	    break;
	case 'l':
	    irq_fn = do_irq_bar;
	    break;
	case 'r':
	    refresh_rate = atoi(optarg) * 1000;
	    break;
	case 's':
	    sysmon = malloc(strlen(optarg) * sizeof(char));
	    strncpy(sysmon, optarg, strlen(optarg));
	    break;
	case 'v':
	    printf("tcsysmon 1.4\n");
	    exit(0);
	    break;
#ifdef TWM
	case 'x':
	    dockapp.x = atoi(optarg);
	    break;
	case 'y':
	    dockapp.y = atoi(optarg);
	    break;
#endif
	}
    }
    if (sysmon == NULL) {
	sysmon = "/dev/sysmon";
    }

    if ((fd = open(sysmon, O_WRONLY)) == -1) {
	sysmon = "/dev/misc/sysmon";
	if ((fd = open(sysmon, O_WRONLY)) == -1) {
	    perror("Cannot open sysmon device");
	    fprintf(stderr, "filename was: %s\n", sysmon);
	    fprintf(stderr, "make sure monitor.o module is loaded\n");
	    fprintf(stderr, "hint: insmod monitor.o\n");
	    exit(-1);
	}
    }
    kernel_check();
    signal(SIGALRM, catch_alarm);

    syscall(__NR_sysinfo, &si);
    dockapp.uptime = si.uptime;
    time(&dockapp.time);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
    /* on my system mem_unit is 4096.  let's be safe and assume it may change,
     * but also assume its a power of 2 (it better be).
     * This only applies to 2.4 kernels anyway. I think this happened after
     * the bigmem patch was merged in.  Total memory was bigger than the
     * long int it was being kept in until then. */
    shift =
	(((((si.
	     mem_unit - 1) - (((si.mem_unit - 1) >> 1) & 0x77777777) -
	    (((si.mem_unit - 1) >> 2) & 0x33333333) -
	    (((si.mem_unit - 1) >> 3) & 0x11111111)) +
	   (((si.
	      mem_unit - 1) - (((si.mem_unit - 1) >> 1) & 0x77777777) -
	     (((si.mem_unit - 1) >> 2) & 0x33333333) -
	     (((si.
		mem_unit -
		1) >> 3) & 0x11111111)) >> 4)) & 0x0F0F0F0F) % 255);
#endif

    /* setup the memory fd */
    if ((memory_fd = open("/proc/meminfo", O_RDONLY)) == -1) {
	perror("Cannot open /proc/meminfo");
	exit(-1);
    }

    /* UI */
    if (!(dockapp.d = XOpenDisplay(dispname))) {
	fprintf(stderr, "Unable to open display '%s'\n", dispname);
	exit(-1);
    }
    new_window("sysmon", 64, 64);
    XFlush(dockapp.d);
    memset(irqpeak, 1, sizeof(irqpeak));

    uptime_thread();
#ifdef PRO
    while (cnt < 64000) {
#else
    while (1) {
#endif
	while (XPending(dockapp.d)) {
	    XNextEvent(dockapp.d, &event);
	    switch (event.type) {
	    case Expose:
		flush_expose();
#ifndef BLACKBOX
		XCopyArea(dockapp.d, dockapp.pixmap, dockapp.iconwin,
			  dockapp.gc, 0, 0, dockapp.width, dockapp.height,
			  0, 0);
#else
		XCopyArea(dockapp.d, dockapp.pixmap, dockapp.win,
			  dockapp.gc, 0, 0, dockapp.width, dockapp.height,
			  0, 0);
#endif
		break;
	    case DestroyNotify:
		XCloseDisplay(dockapp.d);
		exit(0);
		break;
	    }
	}
	draw_thread(irq_fn, memory_fd);
#ifndef PRO
	usleep(refresh_rate);
#else
	cnt++;
#endif
    }
    exit(0);
}
