/* Sysmon - dockapp system monitor
 * Copyright (C) 2000 timecop@japan.co.jp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef MODE5
#ifdef EQ
#error Cannot define MODE5 and EQ mode at the same time!
#endif
#endif
#ifdef EQ
#ifdef HIGHIRQ
#error Cannot define EQ and HIGHIRQ at the same time!
#endif
#endif
#ifdef EQ
#ifdef NORMAL
#error Cannot define NORMAL and EQ mode at the same time!
#endif
#endif
#ifdef MODE5
#ifdef NORMAL
#error Cannot define MODE5 and NORMAL at the same time!
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <pthread.h>
#include <sys/utsname.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <getopt.h>
#include <syscall.h>
#include <errno.h>
#include <linux/kernel_stat.h>
#include <linux/kernel.h>
#include <linux/version.h>

/* x11 includes */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include <X11/xpm.h>

/* pixmaps */
#include "master.xpm"

/* timing */
#ifdef TIMINGS
#include <sys/time.h>
#define TM_INIT()	struct timeval tm_start, tm_end
#define TM_START()	gettimeofday(&tm_start, NULL)
#define TM_END()							 \
do {									 \
    gettimeofday(&tm_end, NULL);					 \
    tm_end.tv_sec  -= tm_start.tv_sec;					 \
    tm_end.tv_usec -= tm_start.tv_usec;					 \
    if (tm_end.tv_usec < 0)						 \
    {									 \
	tm_end.tv_usec += 1000000;					 \
	tm_end.tv_sec  -= 1;						 \
    }									 \
    printf("operation took %ld.%06ld\n", tm_end.tv_sec, tm_end.tv_usec); \
} while(0)
#else
#define TM_INIT()
#define TM_START()
#define TM_END()
#endif

/* UI defines for offsets inside pixmap and things like that. */
#ifdef HIGHIRQ
#define BAR_FULL	5
#define BAR_HALF	3
#else
#define BAR_FULL	6
#define BAR_HALF	3
#endif

/* lights and bars */
#define MODE_RED	0
#define MODE_GREEN	4
#define MODE_OFF	8

/* locations of memory, rio, wio bars. prepare for 24 IRQ setups */
#ifdef HIGHIRQ
#define BAR_MEM_PFY	3, 4
#define BAR_MEM_SWP	3, 13
#define BAR_IO_READ	3, 22
#define BAR_IO_WRIT	3, 25
#else
#define BAR_MEM_PFY	3, 4
#define BAR_MEM_SWP	3, 15
#define BAR_IO_READ	3, 26
#define BAR_IO_WRIT	3, 29
#endif
/* locations of swap and page lights */
#define LGT_PGIN	51, 51
#define LGT_PGOUT	56, 51
#define LGT_SWPIN	51, 56
#define LGT_SWPOUT	56, 56
/* locations of uptime line */
#ifdef HIGHIRQ
#define UPTIME_1	4, 32
#else
#define UPTIME_1	4, 37
#endif

/* sysmon ioctl */
#define SMIOC_GETKSTAT	100
#define SMIOC_GETSI	101	/* unused atm */
#define IRQ_PREP	1	/* nuke the black lights */

/* from fortune */
#define BITCOUNT(x)	(((BX_(x)+(BX_(x)>>4)) & 0x0F0F0F0F) % 255)
#define BX_(x)		((x) - (((x)>>1)&0x77777777)			\
			     - (((x)>>2)&0x33333333)			\
			     - (((x)>>3)&0x11111111))

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
#error Oh my god, you killed Kenny!!! You bastard!
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
#define DK_MAX_DISK DK_NDRIVE
#endif

typedef void (*function) (int *);

typedef struct {
    Display *d;
    int screen;
    Window root;
    Window win;
    Window iconwin;
    Pixmap pixmap;
    Pixmap mask;
    GC gc;
    int width;
    int height;
    time_t uptime;
    time_t time;
} Dockapp;

typedef struct {
    function irq_fn;
    int mem;
} Draw_arg;

/* proto for UI */
static void flush_expose(void);
void new_window(char *, int, int);
static void redraw_window(void);
static void copy_xpm_area(int, int, int, int, int, int);
static void draw_digit(char *string, int x, int y);
static void draw_bar(int x, int y, int width, int mode);
static void draw_light(int mode, int x, int y);
static void draw_peak(int level, int peak, int x, int y, int irq,
		      int *changed);
/* main draw functions */
static void do_irq_bar(int *);
static void do_irq_led(int *);
static void irq_prep(void);

static void draw_thread(void *arg);
static void uptime_thread(void);
static void x11_thread(void);
/* garbage */
void sysmon_help(void);

/* globals */
static Dockapp dockapp;

#ifdef EQ
int pmode[10] = { 92, 90, 88, 86, 84, 82, 80, 78, 76, 74 };
#endif

#if (defined(MODE5) && defined(HIGHIRQ)) || defined(MODE5)
int pmode[5] = { 28, 24, 20, 16, 12 };
#endif

#if (defined(NORMAL) && defined(HIGHIRQ)) || defined(NORMAL) || (defined(HIGHIRQ) && !defined(MODE5))
int pmode[4] = { 24, 20, 16, 12 };
#endif

#if defined(HIGHIRQ)
#define IRQN	24
int xpos[24] =
    { 4, 9, 14, 19, 24, 29, 34, 39, 4, 9, 14, 19, 24, 29, 34, 39, 4, 9, 14,
    19, 24, 29, 34, 39
};
int ypos[24] =
    { 46, 46, 46, 46, 46, 46, 46, 46, 51, 51, 51, 51, 51, 51, 51, 51, 56,
    56, 56, 56, 56, 56, 56, 56
};
#endif

#if defined(EQ)
#define IRQN	16
int xpos[16] =
    { 4, 6, 9, 11, 14, 16, 19, 21, 24, 26, 29, 31, 34, 36, 39, 41 };
int ypos[16] =
    { 51, 51, 51, 51, 51, 51, 51, 51, 51, 51, 51, 51, 51, 51, 51, 51 };
#endif

#if defined(NORMAL) || defined(MODE5) && !defined(HIGHIRQ)
#define IRQN	16
int xpos[16] =
    { 4, 9, 14, 19, 24, 29, 34, 39, 4, 9, 14, 19, 24, 29, 34, 39 };
int ypos[16] =
    { 51, 51, 51, 51, 51, 51, 51, 51, 56, 56, 56, 56, 56, 56, 56, 56 };
#endif

char buf[2048];			/* might as well make it static.  less cycles */
struct kernel_stat kstat;	/* where all the magic happens */
struct sysinfo si;		/* ditto */
unsigned long int irq[IRQN];	/* system IRQ table.. wtf it has 224 intrs? */
unsigned int irqpeak[IRQN];	/* this tries to keep insane peaks from updating */
int fd;				/* sysmon fd for ioctl calls */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
int shift;			/* amount to bitshift to calculate some memory stuff */
#endif
int refresh_rate = 102400;	/* refresh rate. modified by the -r param (100<<10) */

/* UI */
static void flush_expose(void)
{
    XEvent dummy;
    while (XCheckTypedWindowEvent(dockapp.d, dockapp.win, Expose, &dummy));
    while (XCheckTypedWindowEvent
	   (dockapp.d, dockapp.iconwin, Expose, &dummy));
}

void new_window(char *name, int width, int height)
{
    XpmAttributes attr;
    Pixel fg, bg;
    XGCValues gcval;
#ifdef HIGHIRQ
    GC temp;
#endif
    XSizeHints sizehints;
    XClassHint classhint;
    XWMHints wmhints;
    XTextProperty wname;

    dockapp.width = width;
    dockapp.height = height;
    dockapp.screen = DefaultScreen(dockapp.d);
    dockapp.root = DefaultRootWindow(dockapp.d);

    sizehints.flags = USSize | USPosition;
    sizehints.x = 0;
    sizehints.y = 0;
    sizehints.width = width;
    sizehints.height = height;

    fg = BlackPixel(dockapp.d, dockapp.screen);
    bg = WhitePixel(dockapp.d, dockapp.screen);

    dockapp.win = XCreateSimpleWindow(dockapp.d, dockapp.root,
				      sizehints.x, sizehints.y,
				      sizehints.width, sizehints.height, 1,
				      fg, bg);

    dockapp.iconwin =
	XCreateSimpleWindow(dockapp.d, dockapp.win, sizehints.x,
			    sizehints.y, sizehints.width, sizehints.height,
			    1, fg, bg);

    XSetWMNormalHints(dockapp.d, dockapp.win, &sizehints);
    classhint.res_name = name;
    classhint.res_class = name;
    XSetClassHint(dockapp.d, dockapp.win, &classhint);

    XSelectInput(dockapp.d, dockapp.win,
		 ExposureMask | StructureNotifyMask);
    XSelectInput(dockapp.d, dockapp.iconwin,
		 ExposureMask | StructureNotifyMask);

    XStringListToTextProperty(&name, 1, &wname);
    XSetWMName(dockapp.d, dockapp.win, &wname);

    gcval.foreground = fg;
    gcval.background = bg;
    gcval.graphics_exposures = 1;

    dockapp.gc =
	XCreateGC(dockapp.d, dockapp.win,
		  GCForeground | GCBackground | GCGraphicsExposures,
		  &gcval);

    attr.exactColors = 0;
    attr.alloc_close_colors = 1;
    attr.closeness = 30000;
    attr.valuemask = (XpmExactColors | XpmAllocCloseColors | XpmCloseness);
    if (XpmCreatePixmapFromData(dockapp.d, dockapp.root,
				master_xpm, &dockapp.pixmap, &dockapp.mask,
				&attr) != XpmSuccess) {
	fprintf(stderr, "Not enough colors!\n");
	exit(-1);
    }
#ifdef HIGHIRQ
    XCopyArea(dockapp.d, dockapp.pixmap, dockapp.pixmap, dockapp.gc, 64, 0,
	      dockapp.width, dockapp.height, 0, 0);
    temp =
	XCreateGC(dockapp.d, dockapp.mask,
		  GCForeground | GCBackground | GCGraphicsExposures,
		  &gcval);
    XCopyArea(dockapp.d, dockapp.mask, dockapp.mask, temp, 64, 0,
	      dockapp.width, dockapp.height, 0, 0);
#endif

    XShapeCombineMask(dockapp.d, dockapp.win, ShapeBounding, 0, 0,
		      dockapp.mask, ShapeSet);
    XShapeCombineMask(dockapp.d, dockapp.iconwin, ShapeBounding, 0, 0,
		      dockapp.mask, ShapeSet);

    wmhints.initial_state = WithdrawnState;
    wmhints.icon_window = dockapp.iconwin;
    wmhints.icon_x = sizehints.x;
    wmhints.icon_y = sizehints.y;
    wmhints.window_group = dockapp.win;
    wmhints.flags =
	StateHint | IconWindowHint | IconPositionHint | WindowGroupHint;
    XSetWMHints(dockapp.d, dockapp.win, &wmhints);
    XMapWindow(dockapp.d, dockapp.win);
}

static void redraw_window(void)
{
    XCopyArea(dockapp.d, dockapp.pixmap, dockapp.iconwin, dockapp.gc,
	      0, 0, dockapp.width, dockapp.height, 0, 0);
    XCopyArea(dockapp.d, dockapp.pixmap, dockapp.win, dockapp.gc,
	      0, 0, dockapp.width, dockapp.height, 0, 0);
}

static void copy_xpm_area(int x, int y, int w, int h, int dx, int dy)
{
    XCopyArea(dockapp.d, dockapp.pixmap, dockapp.pixmap, dockapp.gc, x, y,
	      w, h, dx, dy);
}

static void draw_digit(char *string, int x, int y)
{
    register int i;
    register int c;
    register int k;

    k = x;
    for (i = 0; string[i]; i++) {
	c = string[i];
	c -= '0';
	copy_xpm_area(c * 6, 64, 6, 8, k, y);
	k += 6;
    }
}

static void draw_bar(int x, int y, int width, int mode)
{
    copy_xpm_area(0, 74, width, mode, x, y);
    copy_xpm_area(0, 81, 58 - width, mode, x + width, y);
}

static void draw_light(int mode, int x, int y)
{
#ifdef EQ
    if (mode > 32) {
	copy_xpm_area(mode, 64, 2, 9, x, y);
    } else {
	copy_xpm_area(mode, 88, 4, 4, x, y);
    }
#else
    copy_xpm_area(mode, 88, 4, 4, x, y);
#endif
}

static void draw_peak(int level, int peak, int x, int y, int irq,
		      int *changed)
{

    /* this function was taken from wmsysmon 0.6.0
     * with minor modifications for usability.
     * see readme for copyright information */
    int a = 0;

    if (peak) {
	if (level) {
#ifdef EQ
	    a = level * 10 / peak;
#endif

#if (defined(MODE5) && defined(HIGHIRQ)) || defined(MODE5)
	    a = (level << 2) / peak;
#endif

#if (defined(NORMAL) || defined(HIGHIRQ)) && !defined(MODE5)
	    a = level * 3 / peak;
#endif
	} else {
	    a = peak;
	}
    }
#ifdef MODE5
    if (a > 4)
	a = 4;
#endif
#if (defined(NORMAL) || defined(HIGHIRQ)) && !defined(MODE5)
    if (a > 3)
	a = 3;
#endif
#ifdef EQ
    if (a > 9)
	a = 9;
#endif
    if (irqpeak[irq] != a) {
	draw_light(pmode[a], x, y);
	*changed = 1;
	irqpeak[irq] = a;
    }
}

static void do_irq_bar(int *changed)
{
    static unsigned long int lints[IRQN];
    static int peak[IRQN];
    int intdiff;
    register int i;

    /* no more irqstat ioctl, this has been provided in kstat all along */
    irq_prep();
    /* draw interrupt bars.  peak calculation code was taken
     * from wmsysmon 0.6.0. see readme for copyright info */
    for (i = 0; i < IRQN; i++) {
	intdiff = irq[i] - lints[i];
	/* did you know the next line divides by 2? */
	peak[i] = (intdiff + peak[i]) >> 1;
	draw_peak(intdiff, peak[i], xpos[i], ypos[i], i, changed);
    }
    memcpy(lints, irq, sizeof(lints));
}

static void do_irq_led(int *changed)
{
    static unsigned long int lints[IRQN];
    static short int lmask[IRQN];
    register int i;

    /* no more irqstat ioctl, this has been provided in kstat all along */
    irq_prep();
    /* draw interrupt lights */
    for (i = 0; i < IRQN; i++) {
	if (irq[i] > lints[i] && !lmask[i]) {
	    draw_light(MODE_GREEN, xpos[i], ypos[i]);
	    lmask[i] = 1;
	    *changed = 1;
	    /* printf("draw IRQ %d ON\n", i); */
	} else if ((irq[i] == lints[i]) && lmask[i]) {
	    draw_light(MODE_OFF, xpos[i], ypos[i]);
	    lmask[i] = 0;
	    *changed = 1;
	    /* printf("draw IRQ %d OFF\n", i); */
	}
    }
    memcpy(lints, irq, sizeof(lints));
}

static void irq_prep(void)
{
    register int i, j;

    memset(irq, 0, sizeof(irq));
    for (j = 0; j < smp_num_cpus; j++) {
	for (i = 0; i < IRQN; i++) {
	    irq[i] += kstat.irqs[cpu_logical_map(j)][i];
	}
    }
}

/* after some thought, moving all the updates into one big function wasn't
 * such a bad idea.  Before the shit was separate because it read from 3
 * different fd's, but now it just does 2 ioctl's so, might as well keep
 * them all here, especially since some call the ioctl more than once.
 * Now this even runs in it's own thread.  For no reason, really. I just
 * wanted to see if I can do it. Of course now I can get rid of all the
 * static vars. */
static void draw_thread(void *arg)	// function irq_fn, int mem)
{
    Draw_arg *draw_arg = (Draw_arg *) arg;
#ifndef EQ
    function irq_fn = draw_arg->irq_fn;
#endif
    int mem = draw_arg->mem;
    /* general */
    int changed;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
    static int count;
#endif

    /* swap, io */
    static unsigned long int oldpin, oldpout, oldsin, oldsout;
    static int p, pp, ppp, pppp;
    static int r, rr;
#ifndef IO_MAJOR
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
    register int major;
#endif
#endif
    register int disk;

    /* interrupts + io */
    long int read_io, write_io;
    static long int old_read_io, old_write_io;

    /* memory */
    char *ptr;
    unsigned long cached;
    unsigned int mbar, sbar;
    long int diff;
    static unsigned int oldmem, oldswap;

    int cnt = 0;

    while (cnt < 64000) {
	/* draw uptime.  but not every 100ms */
	read_io = 0;
	write_io = 0;
	changed = 0;

	/* get system info */
	ioctl(fd, SMIOC_GETKSTAT, &kstat);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	/* this WILL go away */
	syscall(__NR_sysinfo, &si);
#endif

	/* draw IRQ lights or bars */
#ifdef EQ
	/* bars only */
	do_irq_bar(&changed);
#else
	/* choose */
	(*irq_fn) (&changed);
#endif

	/* draw IO and swap */
	/* kernel 2.2 doesnt have anything fancy.  just total'em up. */
#ifndef IO_MAJOR
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	for (disk = 0; disk < DK_MAX_DISK; disk++) {
	    read_io += kstat.dk_drive_rio[disk];
	    write_io += kstat.dk_drive_wio[disk];
	}
#else
	for (major = 0; major < DK_MAX_MAJOR; major++) {
	    for (disk = 0; disk < DK_MAX_DISK; disk++) {
		int active = kstat.dk_drive_rio[major][disk] +
		    kstat.dk_drive_rblk[major][disk] +
		    kstat.dk_drive_wio[major][disk] +
		    kstat.dk_drive_wblk[major][disk];
		if (active)
		    read_io += kstat.dk_drive_rio[major][disk];
		write_io += kstat.dk_drive_wio[major][disk];
	    }
	}
#endif
#else
	for (disk = 0; disk < DK_MAX_DISK; disk++) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	    read_io += kstat.dk_drive_rio[disk];
	    write_io += kstat.dk_drive_wio[disk];
#else
	    read_io += kstat.dk_drive_rio[IO_MAJOR][disk];
	    write_io += kstat.dk_drive_wio[IO_MAJOR][disk];
#endif
	}
#endif

	if (read_io > old_read_io) {
	    diff = read_io - old_read_io;
	    if (diff > 58)
		diff = 58;
	    draw_bar(BAR_IO_READ, diff, BAR_HALF);
	    changed = 1;
	    r = 0;
	    old_read_io = read_io ? read_io : old_read_io;
	} else if (!r) {
	    draw_bar(BAR_IO_READ, 0, BAR_HALF);
	    changed = 1;
	    r = 1;
	}

	if (write_io > old_write_io) {
	    diff = write_io - old_write_io;
	    if (diff > 58)
		diff = 58;
	    draw_bar(BAR_IO_WRIT, diff, BAR_HALF);
	    changed = 1;
	    rr = 0;
	    old_write_io = write_io ? write_io : old_write_io;
	} else if (!rr) {
	    draw_bar(BAR_IO_WRIT, 0, BAR_HALF);
	    changed = 1;
	    rr = 1;
	}

	/* swap, page in/out */
	if (kstat.pgpgin > oldpin) {
	    draw_light(MODE_RED, LGT_PGIN);
	    changed = 1;
	    p = 0;
	    oldpin = kstat.pgpgin;
	} else if (!p) {
	    draw_light(MODE_OFF, LGT_PGIN);
	    changed = 1;
	    p = 1;
	}

	if (kstat.pgpgout > oldpout) {
	    draw_light(MODE_RED, LGT_PGOUT);
	    changed = 1;
	    pp = 0;
	    oldpout = kstat.pgpgout;
	} else if (!pp) {
	    draw_light(MODE_OFF, LGT_PGOUT);
	    changed = 1;
	    pp = 1;
	}

	if (kstat.pswpin != oldsin) {
	    draw_light(MODE_RED, LGT_SWPIN);
	    changed = 1;
	    ppp = 0;
	    oldsin = kstat.pswpin;
	} else if (!ppp) {
	    draw_light(MODE_OFF, LGT_SWPIN);
	    changed = 1;
	    ppp = 1;
	}

	if (kstat.pswpout != oldsout) {
	    draw_light(MODE_RED, LGT_SWPOUT);
	    changed = 1;
	    pppp = 0;
	    oldsout = kstat.pswpout;
	} else if (!pppp) {
	    draw_light(MODE_OFF, LGT_SWPOUT);
	    changed = 1;
	    pppp = 1;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	/* holy shit.  I ran this code on 2.2.10 and it was using fucking 8% CPU.
	 * What actually happens, is under 2.2, si_meminfo walks through large
	 * amounts of memory pages to see if they are reserved or not.  Since
	 * proc driver calls si_meminfo to get memory information, it uses up
	 * quite a bit of CPU time.  So, if we are under 2.2 we'll only update
	 * once every 10 real updates. Swapinfo checking is identical between 
	 * 2.2 and 2.4 so we don't care about that. Also the sysinfo syscall
	 * does a cli() between uptime, nr_procs and load calculations. */
	if (--count < 0) {
	    count = 30;
	    syscall(__NR_sysinfo, &si);
#endif
	    /* memory.  fuck, i wish kernel exported page_cache_size. I wouldn't have
	     * to do this retarded file IO in proc. by the way the stuff in bytes in
	     * proc/meminfo top part is supposed to go away as soon as "all the
	     * tools are updated". so this kB value is all we are going to get
	     * in the future. lets be prepared. */
	    read(mem, &buf, 2047);
	    ptr = strstr(buf, "Cached:");
	    sscanf(ptr, "Cached: %ld", &cached);
	    lseek(mem, 0L, SEEK_SET);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	    /* this crap overflows normal integer operations, so we bit
	     * shift it. */
	    mbar =
		((((si.totalram >> 10) - (si.freeram >> 10)) -
		  (si.bufferram >> 10) -
		  cached) * 58) / (si.totalram >> 10);
#else
	    mbar =
		(((si.totalram - si.freeram) - si.bufferram -
		  ((cached << 10) >> shift)) * 58) / si.totalram;
#endif
	    if (mbar != oldmem) {
		draw_bar(BAR_MEM_PFY, mbar, BAR_FULL);
		changed = 1;
		oldmem = mbar;
	    }
	    /* totalswap is incremented to avoid divide by 0 error. We don't want to
	     * do our math in floats though, so this is what we have to do. error
	     * introduced is within si.mem_unit, on i80x86, 4k. no problem.
	     * if there is any error introduced by this you won't notice until you swap
	     * is at 100% but by that time you got different kind of problems :)
	     * heheheheheheheeh */
	    sbar = ((si.totalswap - si.freeswap) * 58) / ++si.totalswap;
	    if (sbar != oldswap) {
		draw_bar(BAR_MEM_SWP, sbar, BAR_FULL);
		changed = 1;
		oldswap = sbar;
	    }
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	}
#endif
	if (changed) {
	    /* printf("draw_thread: refresh window\n"); */
	    redraw_window();
	}
	usleep(refresh_rate);
	cnt++;
    }
    exit(0);
}

static void uptime_thread(void)
{
    time_t uptime;
    struct tm *upt;
    char buf[10];
    char obuf[10];

    while (1) {

	uptime = time(0) - dockapp.time + dockapp.uptime;

	upt = gmtime(&uptime);
#ifdef _GNU_SOURCE
	snprintf(buf, 10, "%03d:%02d:%02d", upt->tm_yday, upt->tm_hour,
		 upt->tm_min);
#else
	sprintf(buf, "%03d:%02d:%02d", upt->tm_yday, upt->tm_hour,
		upt->tm_min);
#endif

	if (strncmp(buf, obuf, 10)) {
	    draw_digit(buf, UPTIME_1);
	    redraw_window();
	    memcpy(obuf, buf, sizeof(buf));
	}
	sleep(10);
    }
}

/* X11 display event processing thread */
static void x11_thread(void)
{
    XEvent event;
    while (1) {
	while (XPending(dockapp.d)) {
	    XNextEvent(dockapp.d, &event);
	    switch (event.type) {
	    case Expose:
		flush_expose();
		redraw_window();
		break;
	    case DestroyNotify:
		XCloseDisplay(dockapp.d);
		exit(0);
		break;
	    }
	}
	usleep(refresh_rate);
    }
}

void sysmon_help(void)
{
    printf
	("tcsysmon 1.4 - by timecop@japan.co.jp [www.ne.jp/asahi/linux/timecop]\n"
	 "               original concept by dave clark [clarkd@skynet.ca]\n"
	 "               LED bar mode concept by swivel@gnugeneration.com\n"
	 "\n" "command line switches:\n" "\n"
	 "   -d display_name		connect to display_name\n"
#ifndef EQ
	 "   -l				Swivel mode (default off)\n"
#endif
	 "   -r rate			update rate (default 100)\n"
	 "   -h				show this help again\n"
	 "   -s	device			system monitor device file\n"
	 "				( usually /dev/sysmon (see README))\n"
	 "   -v				print version\n");
}

void kernel_check(void)
{

    int r, v, s;
    int r1, v1, s1;

    struct utsname un;

    uname(&un);
    sscanf(un.release, "%d.%d.%d", &r, &v, &s);
    sscanf(UTS_RELEASE, "%d.%d.%d", &r1, &v1, &s1);
    if ((r != r1) || (v != v1)) {
	printf("tcsysmon was compiled for %d.%d kernel.\n"
	       "Recompile is necessary to run it under %d.%d kernel.\n",
	       r1, v1, r, v);
	exit(-1);
    }

}

int main(int argc, char **argv)
{
    char *dispname = NULL;
    char *sysmon = NULL;
    int memory_fd;
    int ch;
    int opts = 0;
    function irq_fn = do_irq_led;
    pthread_t t_draw_thread;
    pthread_t t_x11_thread;
    pthread_t t_uptime_thread;
    Draw_arg draw_arg;

    while ((ch = getopt(argc, argv, "d:hlr:s:v")) != EOF) {
	switch (ch) {
	case 'd':
	    dispname = malloc(strlen(optarg) * sizeof(char));
	    strncpy(dispname, optarg, strlen(optarg));
	    break;
	case 'h':
	    sysmon_help();
	    exit(0);
	    break;
	case 'l':
	    irq_fn = do_irq_bar;
	    opts |= IRQ_PREP;
	    break;
	case 'r':
	    refresh_rate = atoi(optarg) << 10;
	    break;
	case 's':
	    sysmon = malloc(strlen(optarg) * sizeof(char));
	    strncpy(sysmon, optarg, strlen(optarg));
	    break;
	case 'v':
	    printf("tcsysmon 1.4\n");
	    exit(0);
	    break;
	}
    }
    if (sysmon == NULL) {
	sysmon = "/dev/sysmon";
    }

    if ((fd = open(sysmon, O_RDWR)) == -1) {
	perror("cannot open sysmon device");
	fprintf(stderr, "filename was: %s\n", sysmon);
	fprintf(stderr, "make sure monitor.o module is loaded\n");
	fprintf(stderr, "hint: insmod monitor.o\n");
	exit(-1);
    }
    kernel_check();

    syscall(__NR_sysinfo, &si);
    dockapp.uptime = si.uptime;
    dockapp.time = time(0);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
    /* on my system mem_unit is 4096.  let's be safe and assume it may change,
     * but also assume its a power of 2 (it better be).
     * This only applies to 2.4 kernels anyway. I think this happened after
     * the bigmem patch was merged in.  Total memory was bigger than the
     * long int it was being kept in until then. */
    shift = BITCOUNT(si.mem_unit - 1);
#endif

    /* setup the memory fd */
    if ((memory_fd = open("/proc/meminfo", O_RDONLY)) == -1) {
	perror("Cannot open /proc/meminfo");
	exit(-1);
    }

    /* UI */
    if (!(dockapp.d = XOpenDisplay(dispname))) {
	fprintf(stderr, "Unable to open display '%s'\n", dispname);
	exit(-1);
    }
    new_window("sysmon", 64, 64);
    XFlush(dockapp.d);
    if (opts & IRQ_PREP) {
	memset(irqpeak, 1, sizeof(irqpeak));
    }
#ifdef EQ
    memset(irqpeak, 1, sizeof(irqpeak));
#endif

    draw_arg.irq_fn = irq_fn;
    draw_arg.mem = memory_fd;
    pthread_create(&t_draw_thread, NULL, (void *) &draw_thread, &draw_arg);
    pthread_create(&t_uptime_thread, NULL, (void *) &uptime_thread, NULL);
    pthread_create(&t_x11_thread, NULL, (void *) &x11_thread, NULL);
    pthread_exit(NULL);
}
